from django.apps import AppConfig


class DjangoLogLensConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_log_lens'
    label = 'django_log_lens'
    verbose_name = 'Django Log Lens'
